// genexp: force
import 'dart:io';

import 'package:gen_export/src/gen_export.dart';

void generateLibrary({
  required Directory lib,
}) {
  print('[GEN] Library: ${lib.path}');

  if (!lib.existsSync()) throw Exception('lib directory not found');

  final src = Directory(lib.path + Platform.pathSeparator + 'src');
  if (!src.existsSync()) throw Exception('lib/src directory not found');

  final packageName = getPackageName();

  final libFileName =
      lib.absolute.path + '/' + '${packageName}_' + 'library.dart';
  final libFile = File(libFileName);
  if (!libFile.existsSync()) {
    libFile.createSync();
  }

  final library = Library(
    directory: lib,
    name: packageName,
    onlySrc: true,
  );
  library.save();
}

class Library {
  final Directory directory;
  final String? name;
  final List<String> sources = [];
  final List<String> imports = [];
  final List<String> exports = [];
  final bool onlySrc;

  Library({
    required this.directory,
    this.name,
    this.onlySrc = false,
  });

  void _evaluate({
    required Directory directory,
    String? name,
  }) {
    final libFileName = directory.absolute.path +
        '/' +
        (name == null ? '' : '${name}_') +
        'library.dart';

    final libFile = getLibraryFile(libFileName);
    if (libFile == null) return;

    final contents = libFile.readAsStringSync();
    final lines = contents.split('\n');
    for (var line in lines) {
      if (line.trim().startsWith('import ')) {
        final startIdx = contents.indexOf(line);
        final endIdx = contents.indexOf(';', startIdx) + 1;
        final code = contents.substring(startIdx, endIdx);
        imports.add(code);
      }
    }

    final dir = onlySrc ? Directory(directory.path + '/src') : directory;
    addDirectory(dir);
    writeLibFile(libFile);
  }

  void writeLibFile(File libFile) {
    final buffer = StringBuffer();
    buffer.writeln('library $name;');
    buffer.writeln();

    final imps = imports.toSet().toList();
    for (var imp in imps) {
      buffer.writeln(imp);
    }

    buffer.writeln();
    final exps = exports.toSet().toList();
    for (var src in exps) {
      buffer.writeln("export '$src';");
    }

    buffer.writeln();
    final srcs = sources.toSet().toList();
    for (var src in srcs) {
      buffer.writeln("part '$src';");
    }

    libFile.writeAsStringSync(buffer.toString());
  }

  void addDirectory(Directory directory) {
    if (directory != this.directory) {
      final libName = getLibraryName(directory);
      if (libName != null) {
        final library = Library(
          directory: directory,
          name: libName,
        );
        library.save();

        final libFileName = directory.absolute.path +
            '${directory.path.substring(directory.path.lastIndexOf('/'))}_library.dart';

        final src =
            libFileName.replaceAll(this.directory.absolute.path + '/', '');
        exports.add(src);
        return;
      }
    }

    final entities = directory.listSync();
    for (final entity in entities) {
      if (entity is Directory) {
        addDirectory(entity);
        continue;
      }

      if (entity is File) {
        addFile(entity);
      }
    }
  }

  void addFile(File file) {
    if (file.path.toLowerCase().endsWith('.dart')) {
      final libFileName = '${name}_library.dart';
      final filename =
          file.absolute.path.substring(file.absolute.path.lastIndexOf('/') + 1);
      print(filename);
      if (filename == libFileName) {
        return;
      }

      var contents = file.readAsStringSync();

      if (contents.contains('genlib: ignore')) {
        return;
      }

      if (contents.contains('// GENERATED CODE - DO NOT MODIFY BY HAND')) {
        return;
      }

      final lines = contents.split('\n');

      var hasPart = false;
      if (lines.any((element) =>
          element.trim().startsWith('part of') ||
          element.trim().startsWith('part'))) {
        if (name != null && !contents.contains('part of $name')) {
          return;
        }
        hasPart = true;
      }

      for (final line in lines) {
        if (line.trim().startsWith('import ')) {
          final startIdx = contents.indexOf(line);
          final endIdx = contents.indexOf(';', startIdx) + 1;
          final code = contents.substring(startIdx, endIdx);
          contents = contents.replaceAll(code, '');
          imports.add(code);
        }
      }

      if (!hasPart) {
        var finalContent = 'part of $name;\n\n$contents';
        file.writeAsStringSync(finalContent, flush: true);
      } else {
        file.writeAsStringSync(contents, flush: true);
      }

      final src =
          file.absolute.path.replaceAll(directory.absolute.path + '/', '');
      sources.add(src);
    }
  }

  File? getLibraryFile(String name) {
    final libFile = File(name);
    if (libFile.existsSync()) {
      return libFile;
    }

    return null;
  }

  String? getLibraryName(Directory directory) {
    final libFileName = directory.absolute.path +
        '${directory.path.substring(directory.path.lastIndexOf('/'))}_library.dart';

    final libFile = getLibraryFile(libFileName);
    if (libFile == null) return null;

    final contents = libFile.readAsStringSync();
    final idx = contents.indexOf('library');
    final end = contents.indexOf(';', idx);
    final name = contents.substring(idx + 7, end).trim();
    return name;
  }

  void save() {
    _evaluate(
      directory: directory,
      name: name,
    );
  }
}
