// genexp: force
import 'dart:collection';
import 'dart:convert';
import 'dart:io';

import 'package:console/console.dart';
import 'package:translator/translator.dart';
import 'package:yaml/yaml.dart';

void genTranslation() async {
  final pubspecYaml = File('pubspec.yaml');
  if (!pubspecYaml.existsSync()) {
    throw Exception('No pubspec.yaml exists in your current directory');
  }

  final pubspec = loadYaml(pubspecYaml.readAsStringSync());
  final YamlList languages = pubspec['languages'] ?? ['en', 'id'];

  final l10nYaml = File('l10n.yaml');
  if (!l10nYaml.existsSync()) {
    throw Exception('No l10n.yaml exists in your project');
  }

  final translator = GoogleTranslator();

  final config = loadYaml(l10nYaml.readAsStringSync());
  final String arbDir = config['arb-dir'] ?? 'lib/l10n';
  final String outputDir = config['output-dir'] ?? 'lib/localizations';

  final String templateArbFile = config['template-arb-file'] ?? 'app_en.arb';
  var templateName =
      templateArbFile.substring(0, templateArbFile.lastIndexOf('_'));

  removeUnused(File('$arbDir/$templateArbFile'), [outputDir]);

  var defaultLang = templateArbFile.replaceAll('.arb', '');
  defaultLang = defaultLang.substring(defaultLang.lastIndexOf('_') + 1);

  final defaultArb = File('$arbDir/$templateArbFile');
  final defaultArbData =
      json.decode(defaultArb.readAsStringSync()) as Map<String, dynamic>;

  final encoder = JsonEncoder.withIndent('    ');

  for (String lang in languages.where((element) => element != defaultLang)) {
    var countryLang = lang;
    if (lang.contains(RegExp(r'[\-_]'))) {
      countryLang = lang.substring(0, lang.indexOf(RegExp(r'[\-_]')));
    }

    final newArb =
        File('$arbDir/${templateName}_${countryLang.toLowerCase()}.arb');

    print('generate ${newArb.path}');

    var arbData = <String, dynamic>{};
    if (newArb.existsSync()) {
      arbData = json.decode(newArb.readAsStringSync());
    }

    // remove
    arbData.removeWhere((key, value) => !defaultArbData.containsKey(key));

    final progressBar = ProgressBar(complete: defaultArbData.keys.length);
    var progress = 0;
    for (var key in defaultArbData.keys) {
      try {
        if (arbData.containsKey(key)) continue;

        final value = defaultArbData[key];
        if (value is String) {
          final source = value;

          if (source.trim().isEmpty) {
            arbData[key] = source;
            continue;
          }

          try {
            final translation = await translator.translate(
              source,
              from: defaultLang,
              to: lang.toLowerCase().replaceAll('_', '-'),
            );

            arbData[key] = translation.text;
          } catch (e) {
            print('Unable to translate $key to $lang');
          }
        } else {
          print("Can't translate $key, because the value is not a String");
        }
      } finally {
        progress++;
        progressBar.update(progress);
      }
    }

    var sorted = SplayTreeMap.from(arbData);
    newArb.writeAsStringSync(encoder.convert(sorted), flush: true);
  }

  print('Running flutter gen-l10n');
  await Process.run('flutter', ['gen-l10n']);

  print('generation completed!');
}

void removeUnused(File arbFile, List<String> excludes) {
  final Map<String, dynamic> data = json.decode(arbFile.readAsStringSync());

  final lib = Directory('lib');

  final fileContents = lib
      .listSync(recursive: true)
      .whereType<File>()
      .where((element) {
        final path = element.path;
        return !excludes.any((element) => path.startsWith(element));
      })
      .where((element) => element.path.endsWith('.dart'))
      .map((e) => e.readAsStringSync())
      .toList();

  var remove = <String>[];
  for (final key in data.keys) {
    final value = data[key];
    if (value is String) {
      if (fileContents.any((content) => content.contains(key))) {
        continue;
      }

      remove.add(key);
    }
  }

  data.removeWhere((key, value) => remove.contains(key));
  final sorted = SplayTreeMap.from(data);
  final encoder = JsonEncoder.withIndent('    ');
  arbFile.writeAsStringSync(encoder.convert(sorted), flush: true);
}
